/*
QProg.h
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef QPROG_H_
#define QPROG_H_

#define QPROG_VERSION "0.80.1.13"

#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <list>

#include "../../WOK/Text.h"

#include "Algebra/SimpleAlgebra.h"

namespace QProg
{
class Object
{
public:
	Object() : value(NULL) {};
	Object(const std::string& _name, const std::string& _type, void* _value) : name(_name), type(_type), value(_value) {};
	Object(const std::string* _name, const std::string* _type, void* _value) : name(*_name), type(*_type), value(_value) {};

	~Object() {if(name == "temp") cleanup();}

	const std::string& getType() {return type;};
	const std::string& getName() {return name;};

	void* getValue() {return value;};

	void clear()
	{
		value = NULL;
	}
	void cleanup()
	{
		if(value)
		{
			if(type == "string")
			{
				std::string* v = (std::string*) value;
				delete v;
			}
			else if(type == "integer")
			{
				int* v = (int*) value;
				delete v;
			}
			else if(type == "complex")
			{
				Complex* v = (Complex*) value;
				delete v;
			}
			else if(type == "double")
			{
				double* v = (double*) value;
				delete v;
			}
			else if(type == "matrix")
			{
				algebra::mat* v = (algebra::mat*) value;
				delete v;
			}
			else if(type == "vector")
			{
				std::vector <double>* v = (std::vector <double>*) value;
				delete v;
			}
			else if(type == "system")
			{
				std::vector <int>* v = (std::vector <int>*) value;
				delete v;
			}
			else
				throw std::string("Unrecognized type");
		}
		value = NULL;
	}

	void copyTo(Object& newObject)
	{
		newObject.cleanup();
		if(value)
		{
			if(type == "string")
				newObject.value = new std::string(*(std::string*)value);
			else if(type == "integer")
			{
				newObject.value = new int;
				*(int*)newObject.value = *(int*)value;
			}
			else if(type == "complex")
			{
				newObject.value = new Complex;
				*(Complex*)newObject.value = *(Complex*)value;
			}
			else if(type == "double")
			{
				newObject.value = new double;
				*(double*)newObject.value = *(double*)value;
			}
			else if(type == "matrix")
			{
				newObject.value = new algebra::mat;
				*(algebra::mat*)newObject.value = *(algebra::mat*)value;
			}
			else if(type == "system")
			{
				newObject.value = new std::vector <int>;
				*(std::vector <int>*)newObject.value = *(std::vector <int>*)value;
			}
			else
				throw std::string("Unrecognized type");
		}
		value = NULL;
	}

	void print(std::ostream& f, bool printName, bool printAll=false)
	{
		if(!value)
		{
			std::cout << "NULL";
			return;
		}

		if(printName)
			f << type << " " << name << " = ";
		if(type == "string")
			f << *(std::string*)value;
		else if(type == "integer")
			f << *(int*)value;
		else if(type == "complex")
			f << ((Complex*)value)->real << " + " << ((Complex*)value)->imag << "i";
		else if(type == "double")
			f << *(double*)value;
		else if(type == "system")
		{
			for(int i=0; i<(int)((std::vector <int>*)value)->size(); ++i)
				f << (*((std::vector <int>*)value))[i] << ", ";
		}
		else if(type == "vector")
		{
			for(int i=0; i<(int)((std::vector <double>*)value)->size(); ++i)
				f << (*((std::vector <double>*)value))[i] << ", ";
		}
		else if(type == "matrix")
		{
			f << "\n";
			((algebra::mat*)value)->print(f, printAll);
		}
	}

	void setValue(void* v) {value = v;}
	void setName(const std::string& n) {name = n;}
	void setType(const std::string& t) {type = t;}

	void set(Object& input)
	{
		if(!input.value || input.type != type)
			return;

		if(type == "string")
			*(std::string*)value = *(std::string*)input.value;
		else if(type == "integer")
			*(int*)value = *(int*)input.value;
		else if(type == "matrix")
			*(algebra::mat*)value = *(algebra::mat*)input.value;
		else if(type == "complex")
			*(Complex*)value = *(Complex*)input.value;
		else if(type == "double")
			*(double*)value = *(double*)input.value;
		else if(type == "system")
			*(std::vector <int>*)value = *(std::vector <int>*)input.value;
		else if(type == "vector")
			*(std::vector <double>*)value = *(std::vector <double>*)input.value;
		else
			throw std::string("Object::set unkown type");
	}

	void saveToFile(std::ofstream& f);
private:
	std::string name;
	std::string type;
	void* value;
};

class ObjectHelper
{
public:
	static void newString(const std::string& value, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();
		output.setValue(new std::string(value));
		output.setName("temp");
		output.setType("string");
	}

	static void newVector(std::vector <double>& value, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();

		std::vector <double> *v = new std::vector <double>;
		*v = value;

		output.setValue(v);
		output.setName("temp");
		output.setType("vector");
	}

	static void newInteger(int value, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();
		int *i = new int;
		*i = value;
		output.setValue(i);
		output.setName("temp");
		output.setType("integer");
	}

	static void newDouble(double value, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();
		double *i = new double;
		*i = value;
		output.setValue(i);
		output.setName("temp");
		output.setType("double");
	}

	static algebra::mat* newMatrix(int rows, int cols, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();

		algebra::mat *m = new algebra::mat;
		m->create(rows, cols);

		output.setValue(m);
		output.setName("temp");
		output.setType("matrix");

		return m;
	}
	static algebra::mat* newMatrix(algebra::mat* m, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();

		output.setValue(m);
		output.setName("temp");
		output.setType("matrix");

		return m;
	}

	static void newComplex(double real, double imag, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();
		Complex *c = new Complex;
		c->real = real;
		c->imag = imag;
		output.setValue(c);
		output.setName("temp");
		output.setType("complex");
	}
	static void newComplex(Complex i, Object& output)
	{
		if(output.getName() == "temp")
			output.cleanup();
		Complex *c = new Complex;
		*c = i;
		output.setValue(c);
		output.setName("temp");
		output.setType("complex");
	}

	static std::string* getString(Object& input)
	{
		if(!input.getValue() || input.getType() != "string")
			throw std::string("getString error - type is NULL or not string");

		return (std::string*)input.getValue();
	}
	static std::vector <int>* getSystem(Object& input)
	{
		if(!input.getValue() || input.getType() != "system")
		{
			throw std::string("getInteger error - type is NULL or not System");
		}

		return (std::vector <int>*)input.getValue();
	}
	static int* getInteger(Object& input)
	{
		if(!input.getValue() || input.getType() != "integer")
		{
			throw std::string("getInteger error - type is NULL or not int");
		}

		return (int*)input.getValue();
	}
	static double* getDouble(Object& input)
	{
		if(!input.getValue() || input.getType() != "double")
		{
			throw std::string("getInteger error - type is NULL or not double");
		}

		return (double*)input.getValue();
	}
	static Complex* getComplex(Object& input)
	{
		if(!input.getValue() || input.getType() != "complex")
			throw std::string("getComplex error - type is NULL or not complex");

		return (Complex*)input.getValue();
	}
	static std::vector <double>* getVector(Object& input)
	{
		if(!input.getValue() || input.getType() != "vector")
			throw std::string("getComplex error - type is NULL or not vector");

		return (std::vector <double>*)input.getValue();
	}
	static algebra::mat* getMatrix(Object& input)
	{
		if(!input.getValue() || input.getType() != "matrix")
			throw std::string("getMatrix error - type is NULL or not matrix");

		return (algebra::mat*)input.getValue();
	}
};

class State
{
public:
	~State() {cleanup();}
	void cleanup()
	{
		if(!state.empty())
		{
			std::list <Pair>::iterator Iter;
			for(Iter = state.begin(); Iter != state.end(); ++Iter)
			{
				if(!Iter->isCopy)
					Iter->object.cleanup();
			}
		}
		state.clear();
	}
	void set(const std::string& name, Object& input)
	{
		if(state.empty()) return;
		std::list <Pair>::iterator Iter;
		for(Iter = state.begin(); Iter != state.end(); ++Iter)
		{
			if(Iter->object.getName() == name)
				Iter->object.set(input);
		}
	}
	void get(const std::string& name, Object& output)
	{
		if(state.empty()) return;
		std::list <Pair>::iterator Iter;
		for(Iter = state.begin(); Iter != state.end(); ++Iter)
		{
			if(Iter->object.getName() == name)
				output = Iter->object;
		}
	}
	void addState(Object& v)
	{
		state.push_back(Pair(v, false));
	}
	void removeState(std::string& name)
	{
		if(state.empty()) return;
		std::list <Pair> ::iterator Iter;
		for(Iter = state.begin(); Iter != state.end(); ++Iter)
		{
			if(Iter->object.getName() == name)
			{
				if(Iter->isCopy)
					throw std::string("Cannot delete variable");
				Iter->object.cleanup();
				state.erase(Iter);
				return;
			}
		}
	}
	void clear()
	{
		state.clear();
	}
	void copyTo(State& newState)
	{
		newState.cleanup();

		if(!state.empty())
		{
			std::list <Pair>::iterator Iter;
			for(Iter = state.begin(); Iter != state.end(); ++Iter)
			{
				Object newObject;
				Iter->object.copyTo(newObject);	// need to be careful of copies; for now only global is calling this so should be OK
				if(Iter->isCopy)
					throw std::string("Object::copyTo() error; object is a copy");

				newState.state.push_back(Pair(newObject, false));
			}
		}
	}
	void copy(State& newState)
	{
		newState.cleanup();

		if(!state.empty())
		{
			std::list <Pair>::iterator Iter;
			for(Iter = state.begin(); Iter != state.end(); ++Iter)
			{
				newState.state.push_back(Pair(Iter->object, true));
			}
		}
	}

	void print()
	{
		std::cout << "State List:\n";
		if(state.empty()) return;

		std::list <Pair>::iterator Iter;
		for(Iter = state.begin(); Iter != state.end(); ++Iter)
		{
			std::cout << "<" << Iter->object.getType() << ", " << Iter->object.getName() << ">\n";
		}
	}

	void saveToFile(std::ofstream& f);
	bool loadFromFile(std::ofstream& f);
private:
	struct Pair
	{
		Pair() : isCopy(false) {};
		Pair(Object& o, bool _iscopy) : object(o), isCopy(_iscopy) {};

		Object object;
		bool isCopy;
	};
	std::list <Pair> state;
};

class QProg;

class Command
{
public:
	Command(const std::string& name) : commandName(name) {};
	virtual ~Command() {};

	virtual void execute(wok::Text& text, State& state, Object& output, bool globalMode, bool copyState, QProg* qp)=0;
	virtual void printHelp()=0;
	virtual std::string& getName(){return commandName;};
protected:
	std::string commandName;
};

class QProg
{
public:
	~QProg()
	{
		cleanup();
	}
	void cleanup()
	{
		for(unsigned int i=0; i<registeredCommands.size(); ++i)
		{
			if(registeredCommands[i])
				delete (registeredCommands[i]);
		}
		registeredCommands.clear();
	}

	void copyTo(QProg& newQP)
	{
		globalState.copyTo(newQP.globalState);
	}

	void evaluate(const std::string& line, State& state, Object& output, bool globalMode, bool copyState=true);

	void registerCommand(Command* c) {registeredCommands.push_back(c);}

	void preProcess(std::string& text);

	State& getGlobalState() {return globalState;}

	void printGlobal()
	{
		globalState.print();
	}

	void clearGlobalState()
	{
		globalState.cleanup();
	}

	void shell()
	{
		wok::Text text;
		text.addDelimiter(' ');

		Object output;
		int bufferSize = 2048;
		char* buffer = new char[bufferSize];

		std::cout << "Quantum Programmer (Q-Prog) v." << QPROG_VERSION << "\n";
		std::cout << "Copyright (c) 2013; All Rights Reserved\n\n";
		while(1)
		{
			std::cout << "Q-Prog > ";
			if(output.getName() == "temp")
				output.cleanup();
			output.clear();
			std::cin.getline(buffer, bufferSize-2);
			if(buffer[0] == 'e' && buffer[1] == 'x' && buffer[2] == 'i' && buffer[3] == 't' && buffer[4] == 0)
				break;
			if(buffer[0] == '!')
			{
				text.parse(buffer);
				if(text[0] == "!exit")
					break;
				else if(text[0] == "!state")
					printGlobal();
				else if(text[0] == "!clear")
					clearGlobalState();
			}
			else
			{
			try {
				std::string line(buffer);
				preProcess(line);
				evaluate(line, getGlobalState(), output, false);
			}
			catch(std::string& e)
			{
				std::cout << "Error: " << e << "\n";
			}

			output.print(std::cout, true);
			}
			std::cout << "\n\n";
		}

		delete[] buffer;
	}
private:
	State globalState;
	std::vector <Command*> registeredCommands;
};
}
#endif